from datetime import datetime
from json import dumps, loads
from typing import Dict, Optional

try:
    from asyncpg import InterfaceError, create_pool
    from asyncpg.pool import Pool
except ImportError:
    print('asyncpg not installed, PostgresSQL function not available.')
    Pool = None
    InterfaceError = None
    create_pool = None

from minoshiro.enums import Medium, Site
from minoshiro.logger import get_default_logger
from .abc import DataController
from .constants import tables
from .postgres_utils import make_tables, parse_record


class PostgresController(DataController):
    """
    To be able to integrate with an existing database, all tables for minoshiro
    will be put under the `minoshiro` schema unless a different schema name is
    passed to the __init__ method.
    """
    __slots__ = ('pool', 'schema')

    def __init__(self, pool: Pool, logger, schema: str = 'minoshiro'):
        """
        Init method. Create the instance with the `get_instance` method to make
        sure you have all the tables needed.

        :param pool: the `asyncpg` connection pool.

        :param logger: logger object used for logging.

        :param schema: the schema name, default is `minoshiro`
        """
        self.pool = pool
        self.schema = schema
        super().__init__(logger)

    @classmethod
    async def get_instance(cls, logger=None, connect_kwargs: dict = None,
                           pool: Pool = None, schema: str = 'minoshiro'):
        """
        Get a new instance of `PostgresController`

        This method will create the appropriate tables needed.

        :param logger: the logger object.

        :param connect_kwargs:
            Keyword arguments for the
            :func:`asyncpg.connection.connect` function.

        :param pool: an existing connection pool.

        One of `pool` or `connect_kwargs` must not be None.

        :param schema: the schema name used. Defaults to `minoshiro`

        :return: a new instance of `PostgresController`
        """
        logger = logger or get_default_logger()
        assert connect_kwargs or pool, (
            'Please either provide a connection pool or '
            'a dict of connection data for creating a new '
            'connection pool.'
        )
        if not pool:
            try:
                pool = await create_pool(**connect_kwargs)
                logger.info('Connection pool made.')
            except InterfaceError as e:
                logger.error(str(e))
                raise e
        logger.info('Creating tables...')
        await make_tables(pool, schema)
        logger.info('Tables created.')
        return cls(pool, logger, schema)

    def __get_table(self, medium: Medium) -> str:
        """
        Get a table name by medium.

        :param medium: the medium type.

        :return: the table name fot that medium.
        """
        return f'{self.schema}.{tables[medium]}'

    async def get_identifier(self, query: str,
                             medium: Medium) -> Optional[Dict[Site, str]]:
        """
        Get the identifier of a given search query.

        :param query: the search query.

        :param medium: the medium type.

        :return:
            A dict of all identifiers for this search query for all sites,
            None if nothing is found.
        """
        sql = """
        SELECT site, identifier FROM {}.lookup
        WHERE LOWER(syname)=LOWER($1) AND medium=$2;
        """.format(self.schema)

        res = await self.pool.fetch(sql, query, medium.value)
        if not res:
            return
        records = (parse_record(record) for record in res)
        return {Site(site): id_ for site, id_ in records if id_}

    async def set_identifier(self, name: str,
                             medium: Medium, site: Site, identifier: str):
        """
        Set the identifier for a given name.

        :param name: the name.

        :param medium: the medium type.

        :param site: the site.

        :param identifier: the identifier.
        """

        sql = """
        INSERT INTO {}.lookup VALUES ($1, $2, $3, $4)
        ON CONFLICT (syname, medium, site)
        DO UPDATE SET identifier=$4;
        """.format(self.schema)

        await self.pool.execute(sql, name, medium.value,
                                site.value, identifier)

    async def get_mal_title(self, id_: str, medium: Medium) -> Optional[str]:
        """
        Get a MAL title by its id.

        :param id_: th MAL id.

        :param medium: the medium type.

        :return: The MAL title if it's found.
        """
        sql = """
        SELECT title FROM {}.mal
        WHERE id=$1 AND medium=$2;
        """.format(self.schema)
        return await self.pool.fetchval(sql, id_, medium.value)

    async def set_mal_title(self, id_: str, medium: Medium, title: str):
        """
        Set the MAL title for a given id.

        :param id_: the MAL id.

        :param medium: The medium type.

        :param title: The MAL title for the given id.
        """

        sql = """
        INSERT INTO {}.mal VALUES ($1, $2, $3)
        ON CONFLICT (id, medium) DO UPDATE
        SET title=$3;
        """.format(self.schema)

        await self.pool.execute(sql, id_, medium.value, title)

    async def medium_data_by_id(self, id_: str, medium: Medium,
                                site: Site) -> Optional[dict]:
        """
        Get data by id.

        Note that if the data cache is more than 1 day old this will delete
        the row in the DB and return None.

        :param id_: the id.

        :param medium: the medium type.

        :param site: the site.

        :return: the data for that id if found.
        """
        sql = """
        SELECT dict, cachetime FROM {} WHERE id=$1 AND site=$2;
        """.format(self.__get_table(medium))
        res = await self.pool.fetchrow(sql, id_, site.value)
        if not res:
            return
        data, cachetime = parse_record(res)
        if (datetime.now() - cachetime).days < 1:
            return loads(data) if data else None
        else:
            await self.delete_medium_data(id_, medium, site)

    async def set_medium_data(self, id_: str, medium: Medium,
                              site: Site, data: dict):
        """
        Set the data for a given id.

        :param id_: the id.

        :param medium: the medium type.

        :param site: the site.

        :param data: the data for the id.
        """
        sql = """
        INSERT INTO {} VALUES ($1, $2, $3, $4)
        ON CONFLICT (id, site) DO UPDATE
        SET dict=$3, cachetime=$4;
        """.format(self.__get_table(medium))

        await self.pool.execute(
            sql, id_, site.value, dumps(data), datetime.now()
        )

    async def delete_medium_data(self, id_: str, medium: Medium, site: Site):
        """
        Delete a row in medium data table.

        :param id_: the id.

        :param medium: the medium type.

        :param site: the site.
        """
        sql = """
        DELETE FROM {} WHERE id=$1 AND site=$2
        """.format(self.__get_table(medium))

        try:
            await self.pool.execute(sql, id_, site.value)
        except Exception as e:
            self.logger.warning(str(e))



"""
A basic logger
"""
from logging import Formatter, INFO, StreamHandler, getLogger


def get_default_logger():
    """
    Return a basic default logger.

    :return:
        A basic logger with a `StreamHandler` attatched and with level `INFO`
    """
    logger = getLogger('minoshiro')
    console_handler = StreamHandler()
    console_handler.setFormatter(Formatter(
        '%(asctime)s %(levelname)s %(name)s: %(message)s')
    )
    logger.addHandler(console_handler)
    logger.setLevel(INFO)
    return logger



from difflib import SequenceMatcher
from typing import List, Optional

from aiohttp_wrapper import SessionManager

from minoshiro.enums import Medium
from minoshiro.helpers import filter_anime_manga

__escape_table = {
    '&': ' ',
    "\'": "\\'",
    '\"': '\\"',
    '/': ' ',
    '-': ' '
    # '!': '\!'
}

__base_url = 'https://graphql.anilist.co'


def escape(text: str) -> str:
    """
    Escape text for ani list use.

    :param text: the text to be escaped.

    :return: the escaped text.
    """
    return ''.join(__escape_table.get(c, c) for c in text)


def get_closest(query: str, thing_list: List[dict]) -> dict:
    """
    Get the closest matching anime by search query.

    :param query: the search term.

    :param thing_list: a list of animes.

    :return: Closest matching anime by search query if found
                else an empty dict.
    """
    max_ratio, match = 0, None
    matcher = SequenceMatcher(b=query.lower().strip())
    for thing in thing_list:
        ratio = match_max(thing, matcher)
        if ratio > max_ratio and ratio >= 0.90:
            max_ratio = ratio
            match = thing
    return match or {}


def match_max(thing: dict, matcher: SequenceMatcher) -> float:
    """
    Get the max matched ratio for a given thing.

    :param thing: the thing.

    :param matcher: the `SequenceMatcher` with the search query as seq2.

    :return: the max matched ratio.
    """
    thing_name_list = []
    max_ratio = 0
    if 'title' in thing and thing['title'] is not None:
        for title in thing['title']:
            if thing['title'][title]:
                thing_name_list.append(thing['title'][title].lower())

    if 'synonyms' in thing:
        for synonym in thing['synonyms']:
            thing_name_list.append(synonym.lower())

    for name in thing_name_list:
        matcher.set_seq1(name.lower())
        ratio = matcher.ratio()
        if 'one shot' in thing['type'].lower():
            ratio = ratio - .05
        if ratio > max_ratio:
            max_ratio = ratio
    return max_ratio


async def get_entry_by_id(session_manager: SessionManager,
                          medium: Medium, entry_id: str,
                          timeout=3) -> dict:
    """
    Get the full details of an thing by id

    :param session_manager: session manager object

    :param medium: medium to search for

    :param entry_id: thing id.

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: dict with thing info.
    """
    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json',
    }
    data = {
        'query': (__get_query_string(medium, entry_id)).replace('\n', '')
    }
    async with await session_manager.post(
            __base_url, headers=headers, json=data, timeout=timeout) as resp:
        js = await resp.json()

    return js['data']['Media']


async def get_entry_details(session_manager: SessionManager,
                            medium: Medium, query: str,
                            timeout=3) -> Optional[dict]:
    """
    Get the details of an thing by search query.

    :param session_manager: session manager object

    :param medium: medium to search for 'anime', 'manga', 'novel'

    :param query: the search term.

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: dict with thing info.
    """
    if medium not in (Medium.ANIME, Medium.MANGA, Medium.LN):
        raise ValueError('Only Anime, Manga and LN are supported.')
    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json',
    }
    data = {
        'query': f'{__get_query_string(medium, query, True)} }}'
    }
    async with await session_manager.post(
            __base_url, headers=headers, json=data, timeout=timeout) as resp:
        thing = await resp.json()
    closest_entry = get_closest(query, thing['data']['Page']['media'])
    return closest_entry


async def get_page_by_popularity(session_manager, medium: Medium,
                                 page: int, timeout=10) -> Optional[dict]:
    """
    Gets the 40 entries in the medium from specified page.

    :param session_manager: the session manager.

    :param medium: medium 'manga' or 'anime'.

    :param page: page we want info from

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 10.

    :return: dict of page
    """
    med_str = filter_anime_manga(medium)
    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json',
    }
    data = {
        'query': (f'''
            query {{
                Page (page: {page}, perPage: 40) {{
                    media (type: {med_str.upper()} sort: POPULARITY_DESC ) {{
                        title {{
                        romaji
                        english
                        native
                        userPreferred
                        }}
                    synonyms
                    id
                    url: siteUrl
                    type
                    format
                    }}
                }}
                }}''').replace('\n', '')
    }
    async with await session_manager.post(
            __base_url, headers=headers, json=data, timeout=timeout) as resp:
        thing = await resp.json()
    return thing['data']['Page']['media']


def __get_query_string(medium, query, search=False) -> str:
    if medium == Medium.ANIME:
        med_str = 'ANIME'
    else:
        med_str = 'MANGA'
    if search:
        full_str = f'''Page (page: 1, perPage: 40) {{
                media (search: "{query}" type: {med_str})'''
        if medium == Medium.LN:
            full_str = f'''Page (page: 1, perPage: 40) {{
                media (search: "{query}" type: {med_str} format: NOVEL)'''
    else:
        full_str = f'Media (id: {query}, type: {med_str})'
    query = f'''
    query {{
        {full_str} {{
            id
            idMal
            title {{
            romaji
            english
            native
            }}
            url: siteUrl
            startDate {{
            year
            month
            day
            }}
            endDate {{
            year
            month
            day
            }}
            coverImage {{
            large
            medium
            }}
            bannerImage
            format
            type
            status
            episodes
            chapters
            volumes
            season
            description
            averageScore
            meanScore
            genres
            synonyms
            nextAiringEpisode {{
            airingAt
            timeUntilAiring
            episode
            }}
        }}
    }}'''
    return query



"""
Search MAL for anime/manga/lightnovels.
"""

import xml.etree.cElementTree as ET
from difflib import SequenceMatcher
from typing import List, Optional
from urllib.parse import quote

from minoshiro.enums import Medium


async def get_entry_details(session_manager, header_info: dict,
                            medium: Medium, query: str,
                            thing_id: str = None, timeout=3) -> Optional[dict]:
    """
    Get the details of an thing by search query.

    :param session_manager: `SessionManager` instance

    :param medium: the medium type.

    :param header_info: dict containing the header info

    :param query: the search term.

    :param thing_id: thing id.

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: dict with thing info.
    """
    medium_str = 'anime' if medium == Medium.ANIME else 'manga'
    url = (f'https://myanimelist.net/api/{medium_str}/'
           f'search.xml?q={quote(query)}')
    async with await session_manager.get(
            url, headers=header_info, timeout=timeout) as resp:
        html = await resp.text()
        if not html:
            return

    thing_list = []
    for thing in ET.fromstring(html).findall('./entry'):
        synonyms = None
        if thing.find('synonyms').text is not None:
            synonyms = thing.find('synonyms').text.split(";")
        if medium == Medium.ANIME:
            data = {
                'id': thing.find('id').text,
                'url': 'https://myanimelist.net/anime/'
                       f'{thing.find("id").text}',
                'title': thing.find('title').text,
                'english': thing.find('english').text,
                'synonyms': synonyms,
                'episodes': thing.find('episodes').text,
                'type': thing.find('type').text,
                'status': thing.find('status').text,
                'start_date': thing.find('start_date').text,
                'end_date': thing.find('end_date').text,
                'synopsis': thing.find('synopsis').text,
                'image': thing.find('image').text
            }
        else:
            data = {
                'id': thing.find('id').text,
                'url': 'https://myanimelist.net/manga/'
                       f'{thing.find("id").text}',
                'title': thing.find('title').text,
                'english': thing.find('english').text,
                'synonyms': synonyms,
                'chapters': thing.find('chapters').text,
                'volumes': thing.find('volumes').text,
                'type': thing.find('type').text,
                'status': thing.find('status').text,
                'start_date': thing.find('start_date').text,
                'end_date': thing.find('end_date').text,
                'synopsis': thing.find('synopsis').text,
                'image': thing.find('image').text
            }
        thing_list.append(data)

    if thing_id:
        return __get_thing_by_id(thing_id, thing_list)
    else:
        return __get_closest(query.strip(), thing_list)


def __get_closest(query: str, thing_list: List[dict]) -> dict:
    """
    Get the closest matching anime by search query.

    :param query: the search term.

    :param thing_list: a list of anime.

    :return: Closest matching anime by search query if found
                else an empty dict.
    """
    max_ratio, match = 0, None
    matcher = SequenceMatcher(b=query.lower().strip())
    for thing in thing_list:
        ratio = __match_max(thing, matcher)
        if ratio > max_ratio and ratio >= 0.90:
            max_ratio = ratio
            match = thing
    return match or {}


def __match_max(thing: dict, matcher: SequenceMatcher) -> float:
    """
    Get the max matched ratio for a given thing.

    :param thing: the thing.

    :param matcher: the `SequenceMatcher` with the search query as seq2.

    :return: the max matched ratio.
    """
    max_ratio = 0
    matcher.set_seq1(thing['title'].lower())
    ratio = matcher.ratio()
    if ratio > max_ratio:
        max_ratio = ratio
    if thing['synonyms']:
        for synonym in thing['synonyms']:
            matcher.set_seq1(synonym.lower())
            ratio = matcher.ratio()
            if ratio > max_ratio:
                max_ratio = ratio
    return max_ratio


def __get_thing_by_id(thing_id: str,
                      thing_list: List[dict]) -> Optional[dict]:
    """
    Get the max matched ratio for a given thing.

    :param thing_id: the id that we are looking for.

    :param thing_list: the `SequenceMatcher` with the search query as seq2.

    :return: the max matched ratio.
    """
    for thing in thing_list:
        id_ = thing.get('id')
        if str(id_) == thing_id:
            return thing



"""
NovelUpdates.py
Handles all NovelUpdates information
"""
from difflib import SequenceMatcher
from typing import List, Optional
from urllib.parse import quote

from pyquery import PyQuery


async def get_light_novel_url(session_manager, query,
                              names, timeout=3) -> Optional[dict]:
    """
    Get ln url by search query.

    :param session_manager: the `SessionManager` instance.

    :param query: a search query.

    :param names: a list of known names

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: the ln url if it's found.
    """
    params = {
        's': quote(query)
    }

    async with await session_manager.get(
            'http://www.novelupdates.com/?',
            params=params, timeout=timeout) as resp:
        html = await resp.text()

    nu = PyQuery(html)
    ln_list = []

    for thing in nu.find('.w-blog-entry'):
        if PyQuery(thing).find('.w-blog-entry-title').text():
            data = {
                'title': PyQuery(thing).find('.w-blog-entry-title').text(),
                'url': PyQuery(thing).find('.w-blog-entry-link').attr('href')
            }
            ln_list.append(data)
    return __get_closest(query, ln_list, names)


def get_light_novel_by_id(ln_id: str) -> str:
    """
    Returns ln url by id.

    :param ln_id: a ln id.

    :return: the ln url.
    """
    return 'http://novelupdates.com/series/' + str(ln_id)


def __get_closest(query: str, ln_list: List[dict], names) -> dict:
    """
    Get the closest matching light novel by search query.

    :param query: the search term.

    :param ln_list: a list of light novels.

    :param names: a list of known names

    :return:
        Closest matching novel by search query if found else an empty dict.
    """
    synonyms_list = names
    synonyms_list.insert(0, query)
    match = None
    for name in synonyms_list:
        max_ratio = 0
        matcher = SequenceMatcher(b=name.lower())
        for ln in ln_list:
            matcher.set_seq1(ln['title'].lower())
            ratio = matcher.ratio()
            if ratio > max_ratio and ratio >= 0.85:
                max_ratio = ratio
                match = ln
        if match:
            break
    return match or {}



"""
Pull synonyms data from upstream.
"""
from sqlite3 import connect
from time import time
from typing import Tuple

from aiohttp_wrapper import SessionManager

from minoshiro.data import data_path

__all__ = ['get_all_synonyms', 'download_anidb']

__db_path = data_path.joinpath('synonyms.db')
__revision_path = data_path.joinpath('revision')
__anidb_time_path = data_path.joinpath('.anidb_time')
__anidb_xml_path = data_path.joinpath('anime-titles.xml')


async def get_all_synonyms(session_manager) -> list:
    """
    Get all synonyms from the sqlite db.

    :return: all synonyms from the sqlite db.
    """
    await download_db(session_manager)
    with connect(str(__db_path)) as conn:
        cur = conn.cursor()
        cur.execute('SELECT * FROM main.synonyms')
        rows = cur.fetchall()
    return rows


async def download_anidb(session_manager: SessionManager, timestamp=None):
    """

    :param session_manager:
    :param timestamp:
    :return:
    """
    good, new_time = check_anidb_download(timestamp)
    if not good:
        url = 'http://anidb.net/api/anime-titles.xml.gz'
        async with await session_manager.get(url) as resp:
            with __anidb_xml_path.open('wb') as f:
                f.write(await resp.read())
    return good, new_time


async def check_revision(session_manager) -> Tuple[bool, int]:
    """
    Check the revision number from upstream repo.

    :param session_manager: the `SessionManager` instance.

    :return:
        A tuple of (The revision number matched, the upstream revision number)
    """
    url = ('https://raw.githubusercontent.com/Mino-shiro/'
           'minoshiro-database/master/revision')
    async with await session_manager.get(url) as resp:
        revision = int(await resp.text())
    if not __revision_path.is_file():
        return False, revision
    with __revision_path.open() as f:
        local_revision = int(f.read())
    return local_revision == revision, revision


async def download_db(session_manager):
    """
    Download the database from upstream if needed.

    :param session_manager: the `SessionManager` instance.
    """
    good, new = await check_revision(session_manager)
    if good and __db_path.is_file():
        return
    url = ('https://github.com/Mino-shiro/minoshiro-database'
           '/raw/master/synonyms.db')
    async with await session_manager.get(url) as resp:
        content = await resp.read()
    with __db_path.open('wb') as db, __revision_path.open('w+') as rev:
        db.write(content)
        rev.write(str(new))


def check_anidb_download(timestamp=None):
    """

    :param timestamp:

    :return:
    """
    now = int(time())
    if not __db_path.is_file():
        res, new = False, now
    else:
        res, new = check_time(timestamp)
    if res:
        return True, new
    with __anidb_time_path.open('w+') as f:
        f.write(str(now))
    return False, now


def check_time(timestamp=None):
    """
    Check the time for anidb data dump.

    :param timestamp: the current time stamp if any.

    :return:
    """
    now = int(time())
    if timestamp is None:
        if not __anidb_time_path.is_file():
            return False, timestamp
        with __anidb_time_path.open() as f:
            timestamp = int(f.read())
    return now - timestamp < 86400, timestamp



"""
MU.py
Handles all MangaUpdates information
"""
from difflib import SequenceMatcher
from typing import List
from urllib.parse import quote

from pyquery import PyQuery


async def get_manga_url(session_manager, query,
                        names: list, timeout=3) -> dict:
    """
    Get manga url by search query.

    :param session_manager: the `SessionManager` instance.

    :param query: a search query.

    :param names: a list of known names

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: the manga url if it's found.
    """
    params = {
        'search': quote(query)
    }

    async with await session_manager.get(
            'https://mangaupdates.com/series.html',
            params=params, timeout=timeout) as resp:
        html = await resp.text()

    mu = PyQuery(html)
    manga_list = []
    for thing in mu.find('.series_rows_table tr'):
        if PyQuery(thing).find('.col1').text():
            data = {
                'title': PyQuery(thing).find('.col1').text(),
                'url': PyQuery(thing).find('.col1 a').attr('href'),
                'genres': PyQuery(thing).find('.col2').text(),
                'year': PyQuery(thing).find('.col3').text(),
                'rating': PyQuery(thing).find('.col4').text()
            }
            manga_list.append(data)
    return __get_closest(query, manga_list, names)


def get_manga_url_by_id(manga_id) -> str:
    """
    Returns manga url by id.

    :param manga_id: a manga id.

    :return: the manga url.
    """
    return 'https://www.mangaupdates.com/series.html?id=' + str(manga_id)


def __get_closest(query: str, manga_list: List[dict], names) -> dict:
    """
    Get the closest matching manga by search query.

    :param query: the search term.

    :param manga_list: a list of mangas.

    :param names: a list of known names

    :return:
        Closest matching manga by search query if found else an empty dict.
    """
    synonyms_list = names
    synonyms_list.insert(0, query)
    match = None
    for name in synonyms_list:
        max_ratio = 0
        matcher = SequenceMatcher(b=name.lower())
        for manga in manga_list:
            matcher.set_seq1(manga['title'].lower())
            ratio = matcher.ratio()
            if ratio > max_ratio and ratio >= 0.85:
                max_ratio = ratio
                match = manga
        if match:
            break
    return match or {}



"""
Search LNDB for anime.
"""
from difflib import SequenceMatcher
from typing import List, Optional

from aiohttp_wrapper import SessionManager
from pyquery import PyQuery


async def get_light_novel_url(
        session_manager: SessionManager,
        query, names, timeout=3) -> Optional[dict]:
    """
    Get ln url by search query.

    :param session_manager: the `SessionManager` instance.

    :param query: a search query.

    :param names: a list of known names

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: the ln url if it's found.
    """
    query = query.replace(' ', '+')
    params = f'text={query}'

    async with await session_manager.get(
            'http://lndb.info/search?',
            params=params, timeout=timeout) as resp:
        if 'light_novel' in str(resp.url):
            s = (str(resp.url)).rsplit('/', 1)
            title = s[-1].replace('_', ' ')
            return {'title': title, 'url': str(resp.url)}
        html = await resp.text()
    lndb = PyQuery(html)

    ln_list = []
    for thing in lndb.find('#bodylightnovelscontentid table tr'):
        if PyQuery(thing).find('a').text():
            data = {
                'title': PyQuery(thing).find('a').text(),
                'url': PyQuery(thing).find('a').attr('href')
            }
            ln_list.append(data)
    return __get_closest(query, ln_list, names)


def get_light_novel_by_id(ln_id) -> str:
    """
    Returns a light novel url by id.

    :param ln_id: a ln id.

    :return: the ln url
    """
    return 'http://lndb.info/light_novel/' + str(ln_id)


def __get_closest(query: str, ln_list: List[dict], names) -> dict:
    """
    Get the closest matching light novel by search query.

    :param query: the search term.

    :param ln_list: a list of light novels.

    :param names: a list of known names

    :return:
        Closest matching novel by search query if found else an empty dict.
    """
    synonyms_list = names
    synonyms_list.insert(0, query)
    match = None
    for name in synonyms_list:
        max_ratio = 0
        matcher = SequenceMatcher(b=name.lower())
        for ln in ln_list:
            matcher.set_seq1(ln['title'].lower())
            ratio = matcher.ratio()
            if ratio > max_ratio and ratio >= 0.85:
                max_ratio = ratio
                match = ln
        if match:
            break
    return match or {}



"""
Populate the database with some data
before the main search class is initialized.
"""

from aiohttp_wrapper import SessionManager

from .data_controller import DataController
from .enums import Medium, Site
from .helpers import get_synonyms
from .web_api.ani_list import get_page_by_popularity
from .web_api.mal import get_entry_details

__all__ = ['cache_top_pages']


async def cache_top_pages(medium: Medium, session_manager: SessionManager,
                          db: DataController, mal_headers: dict,
                          page_count: int, cache_mal_entries: int, logger):
    """
    Cache the top n pages of anime/manga from Anilist, and try to cache each
    entry for MAL as well.

    :param medium: The medium type.

    :param session_manager: the `SessionManager` instance.

    :param db: the `DataController` instance.

    :param mal_headers: dict of mal auth headers.

    :param page_count: the number of desired pages.

    :param cache_mal_entries: The number of MAL entries to cache.

    :param logger: the logger object.
    """
    assert page_count > 0, 'Please enter a page count greater than 0.'
    await __cache(
        __n_popular_anilist(page_count, medium, session_manager, logger),
        db, medium, logger
    )


async def __cache(async_iter, db, medium, mal_headers, session_manager,
                  cache_mal_entries, logger):
    """
    Cache entries from an `AsyncGenerator`

    :param async_iter: the `AsyncGenerator`

    :param db: the `DataController` instance.

    :param medium: The medium type.

    :param mal_headers: dict of mal auth headers.

    :param session_manager: the `SessionManager` instance.

    :param cache_mal_entries: The number of MAL entries to cache.

    :param logger: the logger object.
    """
    async for entry in async_iter:
        anilist_id = str(entry['id'])
        await db.set_medium_data(anilist_id, medium, Site.ANILIST, entry)
        romanji_name = entry.get('title_romaji')
        english_name = entry.get('title_english')
        anime_name = romanji_name or english_name
        if not anime_name:
            continue
        for syn in get_synonyms(entry, Site.ANILIST):
            await db.set_identifier(syn, medium, Site.ANILIST, anilist_id)


async def __n_popular_anilist(page_count: int, medium: Medium,
                              session_manager: SessionManager, logger):
    """
    Yields top n pages of anime/manga by popularity from Anilist.

    :param page_count: the desired number of pages.

    :param medium: the medium type.

    :param session_manager: the `SessionManager` instance.

    :param logger: the logger object.

    :return: an asynchronous generator that
             yields top n pages of anime/manga for Anilist.
    """
    for i in range(page_count):
        try:
            page_entries = await get_page_by_popularity(
                session_manager, medium, i + 1
            )
            error = False
        except Exception as e:
            logger.warning(f'Error raised by Anilist: {e}')
            page_entries = None
            error = True
        if not page_entries and error:
            continue
        elif not page_entries:
            break
        for entry in page_entries:
            id_ = entry.get('id')
            if id_ or isinstance(id_, int):
                yield entry


async def __cache_anilist_id(name, medium, id_, db):
    """
    Cache an anilist id to the db.

    :param name: the name of the id.

    :param medium: the medium type.

    :param id_: the id.

    :param db: the `DataController` instance.
    """
    if name:
        await db.set_identifier(name, medium, Site.ANILIST, id_)


async def __cache_mal_entry(db, name, medium, mal_headers,
                            session_manager, logger):
    """
    Search MAL from a name, cache the

    :param db: the `DataController` instance.

    :param name: the name of the anime/manga.

    :param medium: the medium type.

    :param mal_headers: the mal auth headers.

    :param session_manager: the `SessionManager` instance.

    :param logger: the logger object.
    """
    id_dict = await db.get_identifier(name, medium) or {}
    mal_id = id_dict.get(Site.MAL)
    try:
        mal_entry = await get_entry_details(
            session_manager, mal_headers, medium, name, mal_id
        )
    except Exception as e:
        logger.debug(f'Error raised by MAL: {e} on item {name}')
        mal_entry = None
    if not mal_entry:
        return

    id_ = mal_entry.get('id')

    if not id_:
        return
    id_ = str(id_)
    title = mal_entry.get('title')
    for syn in get_synonyms(mal_entry, Site.MAL):
        await db.set_identifier(syn, medium, Site.MAL, id_)
    if title:
        await db.set_mal_title(id_, medium, title)
    await db.set_medium_data(id_, medium, Site.MAL, mal_entry)



from minoshiro.enums import Medium

tables = {
    Medium.ANIME: 'anime',
    Medium.MANGA: 'manga',
    Medium.LN: 'ln',
    Medium.VN: 'vn'
}

convert_medium = {
    'Anime': Medium.ANIME,
    'Manga': Medium.MANGA,
    'LN': Medium.LN
}



from sqlite3 import connect


async def make_tables(path, loop):
    """
    Make tables for caching if they don't exist.

    :param path: Path to the database.

    :param loop: the asyncio event loop.
    """
    await loop.run_in_executor(None, __make_tables, path)


def __make_tables(path):
    """
    Make tables for caching if they don't exist.

    :param path: Path to the database.
    """
    with connect(str(path)) as connection:
        connection.execute(
            """
            CREATE TABLE IF NOT EXISTS lookup(
              syname VARCHAR,
              medium INT,
              site INT,
              identifier VARCHAR NOT NULL,
              PRIMARY KEY (syname, medium, site)
            )"""
        )

        connection.execute(
            """
            CREATE TABLE IF NOT EXISTS mal(
              id VARCHAR,
              medium INT,
              title VARCHAR NOT NULL,
              PRIMARY KEY (id, medium)
            )
            """
        )

        tables = """
        CREATE TABLE IF NOT EXISTS {} (
          id VARCHAR,
          site INT,
          dict VARCHAR,
          cachetime INT,
          PRIMARY KEY (id, site)
        )
        """

        for name in ('anime', 'manga', 'ln', 'vn'):
            connection.execute(tables.format(name))
        connection.commit()



"""
Search AniDB for anime.
"""
from difflib import SequenceMatcher
from typing import Dict, Optional

from xmltodict import parse


def process_xml(xml_string: str) -> Dict[str, dict]:
    """
    Process the xml string from the anidb data dump.

    :param xml_string: the xml string.

    :return: A list of dict with keys "id" and "titles".
    """
    parsed = parse(xml_string)
    lst = parsed['animetitles']['anime']
    res = {}
    for anime in (__format_anime(entry) for entry in lst):
        for name in anime['titles']:
            res[name.lower()] = anime
    return res


def get_anime(query: str, anime_list: dict) -> Optional[dict]:
    """
    Get an anime url from a list of animes.

    :param query: the search query.

    :param anime_list: the list of animes.

    :return: the anime id if found, else None.
    """
    anime = anime_list.get(query.lower())
    if anime:
        return anime
    max_ratio, match = 0, None
    matcher = SequenceMatcher(b=query.lower())
    for name, anime in anime_list.items():
        matcher.set_seq1(name.lower())
        ratio = matcher.ratio()
        if ratio > 0.99:
            return anime
        if ratio > max_ratio and ratio >= 0.85:
            max_ratio = ratio
            match = anime
    if match:
        return match


def __format_anime(anime_dict: dict) -> Optional[dict]:
    """
    Format an anime entry from the parsed xml string to a dict.

    :param anime_dict: the input anime dict.

    :return: a dict {"id": the anime id, "titles": the list of titles}
    """
    id_ = anime_dict.get('@aid')
    titles = anime_dict.get('title')
    if not titles or not id_:
        return
    try:
        title_text = [t.get('#text') for t in titles]
    except AttributeError:
        _title = titles.get('#text')
        if not _title:
            return
        title_text = [_title]
    if not title_text:
        return
    return {'id': id_, 'titles': title_text}



from abc import ABCMeta, abstractmethod
from json import loads
from typing import Dict, Optional

from aiohttp_wrapper import SessionManager

from minoshiro.enums import Medium, Site
from minoshiro.upstream import get_all_synonyms
from .constants import convert_medium


class DataController(metaclass=ABCMeta):
    """
    An ABC (abstract base class) that deals with database caching.
    """
    __slots__ = ('logger',)

    def __init__(self, logger):
        """
        :param logger: the logger object to do logging with.
        """
        self.logger = logger

    @abstractmethod
    async def get_identifier(self, query: str,
                             medium: Medium) -> Optional[Dict[Site, str]]:
        """
        Get the identifier of a given search query.

        :param query: the search query.
        :type query: str

        :param medium: the medium type.
        :type medium: Medium

        :return:
            A dict of all identifiers for this search query for all sites,
            None if nothing is found.
        :rtype: Optional[Dict[Site, str]]
        """
        raise NotImplementedError

    @abstractmethod
    async def set_identifier(self, name: str, medium: Medium,
                             site: Site, identifier: str):
        """
        Set the identifier for a given name.

        :param name: the name.
        :type name: str

        :param medium: the medium type.
        :type medium: Medium

        :param site: the site.
        :type site: Site

        :param identifier: the identifier.
        :type identifier: str
        """
        raise NotImplementedError

    @abstractmethod
    async def get_mal_title(self, id_: str, medium: Medium) -> Optional[str]:
        """
        Get a MAL title by its id.

        :param id_: th MAL id.
        :type id_: str

        :param medium: the medium type.
        :type medium: Medium

        :return: The MAL title if it's found.
        :rtype: Optional[str]
        """
        raise NotImplementedError

    @abstractmethod
    async def set_mal_title(self, id_: str, medium: Medium, title: str):
        """
        Set the MAL title for a given id.

        :param id_: the MAL id.
        :type id_: str

        :param medium: The medium type.
        :type medium: Medium

        :param title: The MAL title for the given id.
        :type title: str
        """
        raise NotImplementedError

    @abstractmethod
    async def medium_data_by_id(self, id_: str, medium: Medium,
                                site: Site) -> Optional[dict]:
        """
        Get data by id.

        :param id_: the id.
        :type id_: str

        :param medium: the medium type.
        :type medium: Medium

        :param site: the site.
        :type site: Site

        :return: the data for that id if found.
        :rtype: Optional[dict]
        """
        raise NotImplementedError

    @abstractmethod
    async def set_medium_data(self, id_: str, medium: Medium,
                              site: Site, data: dict):
        """
        Set the data for a given id.

        :param id_: the id.
        :type id_: str

        :param medium: the medium type.
        :type medium: Medium

        :param site: the site.
        :type site: Site

        :param data: the data for the id.
        :type data: dict
        """
        raise NotImplementedError

    async def get_medium_data(self, query: str,
                              medium: Medium) -> Optional[dict]:
        """
        Get the cached data for the given search query.

        :param query: the search query.
        :type query: str

        :param medium: the medium type.
        :type medium: Medium

        :return: the cached data, for all sites that has the data.
        :rtype: Optional[dict]
        """
        id_dict = await self.get_identifier(query, medium)
        if not id_dict:
            return
        return {site: data for site, data in {
            site: await self.medium_data_by_id(id_, medium, site)
            for site, id_ in id_dict.items()
        }.items() if data}

    async def pre_cache(self, session_manager: SessionManager):
        """
        Populate the lookup with synonyms.

        :param session_manager: The Aiohttp SessionManager.
        """
        rows = await get_all_synonyms(session_manager)
        for name, type_, db_links in rows:
            dict_ = loads(db_links)
            mal_name, mal_id = dict_.get('mal', ('', ''))
            anilist = dict_.get('ani')
            ap = dict_.get('ap')
            anidb = dict_.get('adb')
            medium = convert_medium[type_]
            if mal_name and mal_id:
                await self.set_mal_title(str(mal_id), medium, str(mal_name))
            await self.__precache_one(name, medium, Site.MAL, mal_id)
            await self.__precache_one(name, medium, Site.ANILIST, anilist)
            await self.__precache_one(name, medium, Site.ANIMEPLANET, ap)
            await self.__precache_one(name, medium, Site.ANIDB, anidb)

    async def __precache_one(self, name, medium, site, id_):
        if name and (id_ or isinstance(id_, int)):
            await self.set_identifier(str(name), medium, site, str(id_))



"""
Handles all Kitsu api calls
"""
from difflib import SequenceMatcher
from typing import List, Optional
from urllib.parse import quote

from aiohttp_wrapper import SessionManager

from minoshiro.enums import Medium


def get_closest(query: str, thing_list: List[dict]) -> dict:
    """
    Get the closest matching anime by search query.

    :param query: the search term.

    :param thing_list: a list of animes.

    :return: Closest matching anime by search query if found
                else an empty dict.
    """
    max_ratio, match = 0, None
    matcher = SequenceMatcher(b=query.lower().strip())
    for thing in thing_list:
        ratio = match_max(thing, matcher)
        if ratio == 1.0:
            return thing
        if ratio > max_ratio and ratio >= 0.90:
            max_ratio = ratio
            match = thing
    return match


def match_max(thing: dict, matcher: SequenceMatcher) -> float:
    """
    Get the max matched ratio for a given thing.

    :param thing: the thing.

    :param matcher: the `SequenceMatcher` with the search query as seq2.

    :return: the max matched ratio.
    """
    attributes = thing['attributes']
    thing_name_list = []
    max_ratio = 0
    if 'canonicalTitle' in attributes:
        thing_name_list.append(attributes['canonicalTitle'].lower())

    if 'titles' in attributes and attributes['titles'] is not None:
        for title in attributes['titles']:
            if attributes['titles'][title]:
                thing_name_list.append(attributes['titles'][title].lower())

    if attributes.get('abbreviatedTitles'):
        for title in attributes['abbreviatedTitles']:
            thing_name_list.append(title.lower())
    for name in thing_name_list:
        matcher.set_seq1(name.lower())
        ratio = matcher.ratio()
        if 'one shot' in thing['type'].lower():
            ratio = ratio - .05
        if ratio > max_ratio:
            max_ratio = ratio
    return max_ratio


class Kitsu:
    def __init__(self, session_manager: SessionManager, client_id: str,
                 client_secret: str):
        """
        Init the class.
        :param client_id: the Anilist client id.
        :param client_secret: the Anilist client secret.
        """
        self.access_token = None
        # temporary, as application registration isn't implemented yet
        self.client_id = ('dd031b32d2f56c990b1425efe6c42ad847'
                          'e7fe3ab46bf1299f05ecd856bdb7dd')
        self.client_secret = ('54d7307928f63414defd96399fc31ba84'
                              '7961ceaecef3a5fd93144e960c0e151')
        self.session_manager = session_manager
        self.base_url = 'https://kitsu.io/api/edge/'

    async def search_entries(self, medium: Medium,
                             query: str, timeout: int = 3) -> Optional[dict]:
        """
        Get the details of an thing by search query.

        :param medium: medium to search for 'anime', 'manga', 'novel'

        :param query: the search term.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: dict with thing info.
        """
        medium_str = 'anime' if medium == Medium.ANIME else 'manga'
        url = f'{self.base_url}/{medium_str}?filter[text]={quote(query)}'
        headers = {
            'Accept': 'application/vnd.api+json',
            'Content-Type': 'application/vnd.api+json'
        }

        js = await self.session_manager.get_json(
            url, headers=headers, timeout=timeout
        )
        if js:
            closest_entry = get_closest(query, js['data'])
            if closest_entry:
                closest_entry['url'] = (
                    f'https://kitsu.io/{medium_str}/'
                    f'{closest_entry["attributes"]["slug"]}'
                )
            return closest_entry

    async def get_entry_by_id(self, medium, id_, timeout=3) -> Optional[dict]:
        """
        Get the details of a thing by id.

        :param medium: medium to search

        :param id_: the id.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: dict with thing info.
        """
        medium_str = 'anime' if medium == Medium.ANIME else 'manga'
        url = f'{self.base_url}/{medium_str}?filter[slug]={id_}'
        headers = {
            'Accept': 'application/vnd.api+json',
            'Content-Type': 'application/vnd.api+json'
        }

        js = await self.session_manager.get_json(
            url, headers=headers, timeout=timeout
        )

        first = js['data']
        if first[0]:
            first[0]['url'] = first[0]['links']['self']
        return first.pop()



from asyncio import get_event_loop
from itertools import chain
from pathlib import Path
from traceback import format_exc
from typing import Dict, Iterable, Union

from aiohttp_wrapper import SessionManager

from .data import data_path
from .data_controller import (DataController, PostgresController,
                              SqliteController)
from .enums import Medium, Site
from .helpers import get_synonyms
from .logger import get_default_logger
from .pre_cache import cache_top_pages
from .upstream import download_anidb
from .web_api import ani_db, ani_list, anime_planet, kitsu, lndb, mu, nu

from warnings import warn


class Minoshiro:
    def __init__(self, db_controller: DataController,
                 *, logger=None, loop=None):
        """
        Represents the search instance.

        It is suggested to use one of the class methods to create the instance
        if you wish to use one of the data controllers provided by the library.

        Make sure you run the ``pre_cache`` method if you initialized the class
        directly from the ``__init__`` method.

        :param db_controller:
            Any sub class of ``minoshiro.data_controller.abc.DataController``
            will work here.

            It may also contain a key "description" for the description you
            wish to use in the auth header.

            If this key is not present, the description defaults to:
            ``A Python library for anime search.``

        :param logger:
            The logger object. If it's not provided, will use the
            defualt logger provided by the library.

        :param loop:
            An asyncio event loop. If not provided will use the default
            event loop.
        """
        self.session_manager = SessionManager()

        self.db_controller = db_controller

        self.kitsu = kitsu.Kitsu(
            self.session_manager, '', ''
        )

        self.loop = loop or get_event_loop()
        self.logger = logger or get_default_logger()

        self.__anidb_list = None
        self.__anidb_time = None

    @classmethod
    async def from_postgres(cls, db_config: dict = None,
                            pool=None, *, schema='minoshiro',
                            cache_pages: int = 0,
                            logger=None, loop=None):
        """
        Get an instance of `minoshiro` with class `PostgresController` as the
        database controller.

            It may also contain a key "description" for the description you
            wish to use in the auth header.

            If this key is not present, the description defaults to:
                "A Python library for anime search."

        :param db_config:
            A dict of database config for the connection.

            It should contain the keys in  keyword arguments for the :func:
            `asyncpg.connection.connect` function.

        :param pool: an existing connection pool.

        One of ``db_config`` or ``pool`` must not be None.

        :param schema: the schema name used. Defaults to `minoshiro`

        :param cache_pages:
            The number of pages of anime and manga from Anilist to cache
            before the instance is created. Each page contains 40 entries max.

        :param logger:
            The logger object. If it's not provided, will use the
            defualt logger provided by the library.

        :param loop:
            An asyncio event loop. If not provided will use the default
            event loop.

        :return:
            Instance of `minoshiro` with class `PostgresController`
            as the database controller.
        """
        assert db_config or pool, (
            'Please either provide a connection pool or '
            'a dict of connection data for creating a new '
            'connection pool.'
        )
        logger = logger or get_default_logger()
        db_controller = await PostgresController.get_instance(
            logger, db_config, pool, schema=schema
        )
        instance = cls(db_controller, logger=logger, loop=loop)
        await instance.pre_cache(cache_pages)
        return instance

    @classmethod
    async def from_sqlite(cls, path: Union[str, Path], *,
                          cache_pages: int = 0,
                          logger=None, loop=None):
        """
        Get an instance of `minoshiro` with class `SqliteController` as the
        database controller.


            It may also contain a key "description" for the description you
            wish to use in the auth header.

            If this key is not present, the description defaults to:
                "A Python library for anime search."

        :param path:
            The path to the SQLite3 database, can either be a string or a
            Pathlib Path object.

        :param cache_pages:
            The number of pages of anime and manga from Anilist to cache
            before the instance is created. Each page contains 40 entries max.

        :param logger:
            The logger object. If it's not provided, will use the
            defualt logger provided by the library.

        :param loop:
            An asyncio event loop. If not provided will use the default
            event loop.

        :return:
            Instance of `minoshiro` with class `PostgresController`
            as the database controller.
        """
        logger = logger or get_default_logger()
        db_controller = await SqliteController.get_instance(path, logger, loop)
        instance = cls(db_controller,
                       logger=logger, loop=loop)
        await instance.pre_cache(cache_pages)
        return instance

    async def pre_cache(self, cache_pages: int):
        """
        Pre cache the database with anime and managa data.

        :param cache_pages:
            Number of Anilist pages to cache. There are 40 entries per page.

        """
        assert cache_pages >= 0, 'Param `cache_pages` must not be negative.'
        self.logger.info('Populating lookup...')
        await self.db_controller.pre_cache(self.session_manager)
        self.logger.info('Lookup populated.')

        self.logger.info('Populating data...')

        for med in (Medium.ANIME, Medium.MANGA):
            if cache_pages:
                await cache_top_pages(
                    med, self.session_manager, self.db_controller,
                    cache_pages, self.logger
                )

        self.logger.info('Data populated.')
        await self.__fetch_anidb()

    async def yield_data(self, query: str, medium: Medium,
                         sites: Iterable[Site] = None, *, timeout=3):
        """
        Yield the data for the search query from all sites.

        :param query: the search query.

        :param medium: the medium type.

        :param sites:
            an iterable of sites desired. If None is provided, will
            search all sites by default.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return:
            an asynchronous generator that yields the site and data
            in a tuple for all sites requested.
        """
        sites = sites if sites else list(Site)
        cached_data, cached_id = await self._get_cached(query, medium)
        to_be_cached = {}
        names = []
        for site in sites:
            res, id_ = await self._get_result(
                cached_data, cached_id, query, names, site, medium, timeout
            )
            if res:
                yield site, res
                for title in get_synonyms(res, site):
                    names.append(title)
            if id_:
                to_be_cached[site] = id_
        await self._cache(to_be_cached, names, medium)

    async def get_data(self, query: str, medium: Medium,
                       sites: Iterable[Site] = None, *,
                       timeout=3) -> Dict[Site, dict]:
        """
        Get the data for the search query in a dict.

        :param query: the search query.

        :param medium: the medium type.

        :param sites:
            an iterable of sites desired. If None is provided, will
            search all sites by default.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: Data for all sites in a dict {Site: data}
        """
        return {site: val async for site, val in self.yield_data(
            query, medium, sites, timeout=timeout
        )}

    async def _cache(self, to_be_cached, names, medium):
        """
        Cache search results into the db.

        :param to_be_cached: items to be cached.

        :param names: all names for the item.

        :param medium: the medium type.
        """
        itere = set(chain(*names))
        for site, id_ in to_be_cached.items():
            await self.__cache_one(site, id_, medium, itere)

    async def __cache_one(self, site, id_, medium, iterator):
        """
        Cache one id.

        :param site: the site.

        :param id_: the id.

        :param medium: the medium type.

        :param iterator: an iterator for all names.
        """
        for name in iterator:
            if name:
                await self.db_controller.set_identifier(
                    name, medium, site, id_
                )

    async def __fetch_anidb(self):
        """
        Fetch data dump from anidb if one of the following is True:
            The data dump file is not found.
            The data dump file is more than a day old.
        """
        dump_path = data_path.joinpath('anime-titles.xml')
        self.logger.info('Checking anidb conditions...')
        good, new_time = await download_anidb(
            self.session_manager, self.__anidb_time
        )
        if good:
            self.logger.info(
                'Anidb condition is good, no new data downloaded.'
            )
        else:
            self.logger.info(
                'Anidb condition is outdated, new data downloaded.'
            )

        if not good or not self.__anidb_list:
            self.logger.info('Reading anidb data from disk...')
            with dump_path.open() as xml_file:
                xml = xml_file.read()
            self.__anidb_list = ani_db.process_xml(xml)
            self.logger.info('Anidb data read from disk.')
        self.__anidb_time = new_time

    async def __find_anilist(self, cached_data, cached_ids,
                             medium, query, timeout):
        """
        Find Anilist data.

        Return the cached data if it exists.

        If there are no cached data, attempt make an api call to Anilist and
        find the data. Return and cache the api call result if it's found,
        else return None.

        :param cached_data: a dict of cached data.

        :param cached_ids: a dict of cached ids.

        :param medium: the medium type.

        :param query: the search query.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: the anilist data and id in a tuple if found.
        """
        if medium not in (Medium.ANIME, Medium.MANGA, Medium.LN):
            return None, None
        cached_anilist = cached_data.get(Site.ANILIST)
        if cached_anilist:
            return cached_anilist, str(cached_anilist['id'])

        anilist_id = cached_ids.get(Site.ANILIST) if cached_ids else None

        if anilist_id:
            resp = await ani_list.get_entry_by_id(
                self.session_manager, medium, anilist_id, timeout
            )
        else:
            resp = await ani_list.get_entry_details(
                self.session_manager, medium, query, timeout
            )

        id_ = str(resp['id']) if resp else None
        try:
            return resp, id_
        finally:
            if resp and id_:
                await self.db_controller.set_medium_data(
                    id_, medium, Site.ANILIST, resp
                )

    # Deprecated
    async def __find_mal(self, cached_data, cached_ids,
                         medium, query, timeout):
        """
        Find MAL data.

        Return the cached data if it exists.

        If there are no cached data, attempt make an api call to MAL and
        find the data. Return and cache the api call result if it's found,
        else return None.

        :param cached_data: a dict of cached data.

        :param cached_ids: a dict of cached ids.

        :param medium: the medium type.

        :param query: the search query.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: the MAL data and id in a tuple  if found.
        """
        """
        if medium not in (Medium.ANIME, Medium.MANGA, Medium.LN):
            return None, None
        cached_mal = cached_data.get(Site.MAL)
        if cached_mal:
            return cached_mal, str(cached_mal['id'])
        mal_id = cached_ids.get(Site.MAL) if cached_ids else None
        if mal_id:
            cached_title = await self.db_controller.get_mal_title(
                mal_id, medium)
            if cached_title:
                query = cached_title

        resp = await mal.get_entry_details(
            self.session_manager, self.mal_headers, medium, query, mal_id,
            timeout
        )

        id_ = str(resp['id']) if resp else None
        try:
            return resp, id_
        finally:
            if resp and id_:
                await self.db_controller.set_mal_title(
                    id_, medium, resp['title']
                )
                await self.db_controller.set_medium_data(
                    id_, medium, Site.MAL, resp
                )
        """

    async def __find_anidb(self, cached_ids, medium, query):
        """
        Find Anidb url.

        Return the cached url if it's found.

        If no cached url is found, try search through the datadump,
        return and cache the url if it's found.

        :param cached_ids: a dict of cached ids.

        :param query: the search query.

        :return: The data and id in a tuple if found.
        """
        if medium != Medium.ANIME:
            return None, None
        cached_id = cached_ids.get(Site.ANIDB) if cached_ids else None
        base_url = 'https://anidb.net/perl-bin/animedb.pl?show=anime&aid='
        if cached_id:
            return {'url': f'{base_url}{cached_id}'}, cached_id
        await self.__fetch_anidb()
        res = await self.loop.run_in_executor(
            None, ani_db.get_anime, query, self.__anidb_list
        )
        if not res:
            return None, None
        id_ = res['id']
        res['url'] = f'{base_url}{id_}'
        return res, id_

    async def __find_ani_planet(self, cached_ids, medium: Medium,
                                query: str, names: list, timeout):
        """
        Find a anime or manga url from ani planet.

        :param medium: the medium type.

        :param query: the search query.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: the ani planet data and id in a tuple  if found.
        """
        ap_id = cached_ids.get(Site.ANIMEPLANET) if cached_ids else None

        if medium == Medium.ANIME:
            if ap_id:
                return {'url': anime_planet.get_anime_url_by_id(
                    cached_ids.get(Site.ANIMEPLANET))}, None
            else:
                return {'url': await anime_planet.get_anime_url(
                    self.session_manager, query, names, timeout=timeout
                )}, None

        if medium == Medium.MANGA:
            if ap_id:
                return {'url': anime_planet.get_manga_url_by_id(
                    cached_ids.get(Site.ANIMEPLANET))}, None
            else:
                return {'url': await anime_planet.get_manga_url(
                    self.session_manager, query, names, timeout=timeout
                )}, None

        return None, None

    async def __find_kitsu(self, cached_ids, medium, query, timeout):
        """
        Find a anime or manga url from ani planet.

        :param medium: the medium type.

        :param query: the search query.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: the ani planet data and id in a tuple  if found.
        """
        if medium not in (Medium.ANIME, Medium.MANGA, Medium.LN):
            return None, None
        kitsu_id = cached_ids.get(Site.KITSU) if cached_ids else None
        if kitsu_id:
            resp = await self.kitsu.get_entry_by_id(
                medium, kitsu_id, timeout
            )
        else:
            resp = await self.kitsu.search_entries(
                medium, query, timeout
            )
        id_ = str(resp['id']) if resp else None
        return resp, id_

    async def __find_manga_updates(self, cached_ids, medium,
                                   query, names, timeout):
        """
        Find a manga updates url.

        :param medium: the medium type.

        :param query: the search query.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: the data and id in a tuple if found.
        """
        mu_id = cached_ids.get(Site.MANGAUPDATES) if cached_ids else None
        if medium == Medium.MANGA:
            if mu_id:
                return {'url': mu.get_manga_url_by_id(
                    mu_id
                )}, None
            else:
                return await mu.get_manga_url(
                    self.session_manager, query, names, timeout
                ), None

        return None, None

    async def __find_lndb(self, cached_ids, medium, query, names, timeout):
        """
        Find an lndb url.

        :param medium: the medium type.

        :param query: the search query.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: the lndb data and id in a tuple if found.
        """
        lndb_id = cached_ids.get(Site.LNDB) if cached_ids else None
        if medium == Medium.LN:
            if lndb_id:
                return {'url': lndb.get_light_novel_by_id(
                    lndb_id
                )}, None
            else:
                return await lndb.get_light_novel_url(
                    self.session_manager, query, names, timeout), None
        return None, None

    async def __find_novel_updates(self, cached_ids, medium,
                                   query, names, timeout):
        """
        Find a Novel Updates url.

        :param medium: the medium type.

        :param query: the search query.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: the data and id in a tuple if found.
        """
        nu_id = cached_ids.get(Site.NOVELUPDATES) if cached_ids else None
        if medium == Medium.LN:
            if nu_id:
                return {'url': nu.get_light_novel_by_id(
                    nu_id
                )}, None
            return await nu.get_light_novel_url(
                self.session_manager, query, names, timeout
            ), None

        return None, None

    async def __find_vndb(self, medium, query):
        pass

    async def _get_cached(self, query: str, medium: Medium) -> tuple:
        """
        Get cached data from the database.

        :param query: the search query.

        :param medium: the medium type.

        :return: a tuple of (cached data, cached ids)
        """
        identifiers = await self.db_controller.get_identifier(query, medium)
        if not identifiers:
            return {}, None
        entry_resp = {}
        for site, id_ in identifiers.items():
            medium_data = await self.db_controller.medium_data_by_id(
                id_, medium, site)
            if medium_data:
                entry_resp[site] = medium_data

        return entry_resp, identifiers

    async def _get_result(self, cached_data, cached_id, query, names,
                          site: Site, medium: Medium, timeout) -> tuple:
        """
        Get results from a site.

        :param cached_data: the cached data.

        :param cached_id: the cached id.

        :param site: the site.

        :param query: the search query.

        :param medium: the medium type.

        :param timeout:
            The timeout in seconds for each HTTP request. Defualt is 3.

        :return: Search results data and id in a tuple for that site.
        """
        try:
            if site == Site.ANILIST:
                return await self.__find_anilist(
                    cached_data, cached_id, medium, query, timeout
                )

            if site == Site.KITSU:
                return await self.__find_kitsu(
                    cached_id, medium, query, timeout
                )

            if site == site.MAL:
                # 2018-10-05 Deprecating MAL
                warn(
                    "MyAnimeList functionality has been deprecated.",
                    DeprecationWarning,
                    stacklevel=2
                )
                pass

            if site == Site.ANIDB:
                return await self.__find_anidb(cached_id, medium, query)

            if site == Site.ANIMEPLANET:
                return await self.__find_ani_planet(
                    cached_id, medium, query, names, timeout
                )

            if site == Site.MANGAUPDATES:
                return await self.__find_manga_updates(
                    cached_id, medium, query, names, timeout
                )

            if site == Site.LNDB:
                return await self.__find_lndb(
                    cached_id, medium, query, names, timeout
                )

            if site == Site.NOVELUPDATES:
                return await self.__find_novel_updates(
                    cached_id, medium, query, names, timeout
                )

            if site == Site.VNDB:
                return None, None
        except Exception as e:
            self.logger.warning(
                f'Error raised when retriving data from {site}: {e}\n'
                f'{format_exc()}'
            )
            return None, None



from logging import NullHandler, getLogger

from .data_controller import (DataController, PostgresController,
                              SqliteController)
from .enums import Medium, Site
from .logger import get_default_logger
from .minoshiro import Minoshiro

__all__ = ['DataController', 'PostgresController', 'SqliteController',
           'get_default_logger', 'Site', 'Medium', 'Minoshiro']

getLogger(__name__).addHandler(NullHandler())



"""
Database utility functions.
"""

from typing import Optional

try:
    from asyncpg import Record
    from asyncpg.pool import Pool
except ImportError:
    Record = None
    Pool = None
    print('asyncpg not installed, PostgresSQL function not available.')


def parse_record(record: Record) -> Optional[tuple]:
    """
    Parse a asyncpg Record object to a tuple of values

    :param record: the asyncpg Record object

    :return: the tuple of values if it's not None, else None
    """
    try:
        return tuple(record.values())
    except AttributeError:
        return None


async def make_tables(pool: Pool, schema: str):
    """
    Make tables used for caching if they don't exist.

    :param pool: the connection pool.

    :param schema: the schema name.
    """
    await pool.execute('CREATE SCHEMA IF NOT EXISTS {};'.format(schema))

    lookup = """
    CREATE TABLE IF NOT EXISTS {}.lookup (
      syname VARCHAR,
      medium SMALLINT,
      site SMALLINT,
      identifier VARCHAR NOT NULL,
      PRIMARY KEY (syname, medium, site)
    );""".format(schema)

    mal = """
    CREATE TABLE IF NOT EXISTS {}.mal (
      id VARCHAR,
      medium SMALLINT,
      title VARCHAR NOT NULL,
      PRIMARY KEY (id, medium)
    );
    """.format(schema)

    tables = """
    CREATE TABLE IF NOT EXISTS {} (
      id VARCHAR,
      site SMALLINT,
      dict VARCHAR,
      cachetime TIMESTAMP,
      PRIMARY KEY (id, site)
    )
    """
    await pool.execute(lookup)
    await pool.execute(mal)
    for name in ('anime', 'manga', 'ln', 'vn'):
        await pool.execute(tables.format(f'{schema}.{name}'))



"""
Retrieve anime info from AnimePlanet.
"""
from collections import deque
from difflib import SequenceMatcher
from itertools import chain
from typing import List, Optional
from urllib.parse import quote

from pyquery import PyQuery


def sanitize_search_text(text: str) -> str:
    """
    Sanitize text for Anime-Planet use.
    :param text: the text to be escaped.
    :return: the escaped text.
    """
    return text.replace('(TV)', 'TV')


async def get_anime_url(session_manager, query, names: list,
                        timeout=3) -> Optional[str]:
    """
    Get anime url by search query.

    :param session_manager: the `SessionManager` instance.

    :param query: a search query.

    :param names: a list of synonyms

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: the anime url if it's found.
    """
    query = sanitize_search_text(query)
    params = {
        'name': quote(query)
    }
    async with await session_manager.get(
            "http://www.anime-planet.com/anime/all?",
            params=params, timeout=timeout) as resp:
        html = await resp.text()
    ap = PyQuery(html)
    if ap.find('.cardDeck.pure-g.cd-narrow[data-type="anime"]'):
        anime_list = []
        for entry in ap.find('.card.pure-1-6'):
            anime = {
                'title': PyQuery(entry).find('h4').text(),
                'url': (f'http://www.anime-planet.com'
                        f'{PyQuery(entry).find("a").attr("href")}')
            }
            anime_list.append(anime)
        return __get_closest(query, anime_list, names).get('url')
    return ap.find("meta[property='og:url']").attr('content')


async def get_manga_url(session_manager, query,
                        names: list, author_name=None,
                        timeout=3) -> Optional[str]:
    """
    Get manga url by search query.

    :param session_manager: the `SessionManager` instance.

    :param query: a search query.

    :param names: a list of known names

    :param author_name: name for the manga author name

    :param timeout:
        The timeout in seconds for each HTTP request. Defualt is 3.

    :return: the anime url if it's found.
    """
    params = {
        'name': quote(query)
    }
    if author_name:
        params['author'] = quote(author_name)
        async with await session_manager.get(
                "http://www.anime-planet.com/manga/all?",
                params=params, timeout=timeout) as resp:
            html = await resp.text()
        if "No results found" in html:
            rearranged_author_names = deque(
                author_name.split(' '))
            rearranged_author_names.rotate(-1)
            rearranged_name = ' '.join(rearranged_author_names)
            params['author'] = quote(rearranged_name)
            async with await session_manager.get(
                    "http://www.anime-planet.com/manga/all?",
                    params=params, timeout=timeout) as resp:
                html = await resp.text()
    else:
        async with await session_manager.get(
                "http://www.anime-planet.com/manga/all?",
                params=params, timeout=timeout) as resp:
            html = await resp.text()
    ap = PyQuery(html)

    if ap.find('.cardDeck.pure-g.cd-narrow[data-type="manga"]'):
        manga_list = []
        for entry in ap.find('.card.pure-1-6'):
            anime = {
                'title': PyQuery(entry).find('h4').text(),
                'url': (f'http://www.anime-planet.com'
                        f'{PyQuery(entry).find("a").attr("href")}')
            }
            manga_list.append(anime)

            if author_name:
                author_name = author_name.lower().split(' ')
                for manga in manga_list:
                    manga['title'] = manga['title'].lower()
                    for name in author_name:
                        manga['title'] = manga['title'].replace(name, '')
                    manga['title'] = manga['title'].replace('(', '')
                    manga['title'] = manga['title'].replace(')', '').strip()

        return __get_closest(query, manga_list, names).get('url')
    else:
        return ap.find("meta[property='og:url']").attr('content')


def get_anime_url_by_id(anime_id) -> str:
    """
    Returns anime url by id.

    :param anime_id: an anime id.

    :return: the anime url.
    """
    return 'http://www.anime-planet.com/anime/' + str(anime_id)


def get_manga_url_by_id(manga_id) -> str:
    """
    Returns manga url by id.

    :param manga_id: a manga id.

    :return: the manga url.
    """
    return 'http://www.anime-planet.com/manga/' + str(manga_id)


def __get_closest(query: str, anime_list: List[dict], names) -> dict:
    """
    Get the closest matching anime by search query.

    :param query: the search term.

    :param anime_list: a list of anime.

    :param names: a list of known names

    :return:
        Closest matching anime by search query if found else an empty dict.
    """
    for name in chain((query,), names):
        matcher = SequenceMatcher(b=name.lower())
        for anime in anime_list:
            matcher.set_seq1(anime['title'].lower())
            ratio = matcher.ratio()
            if ratio >= 0.85:
                return anime
    return {}



from .enums import Medium, Site


def filter_anime_manga(medium: Medium) -> str:
    if medium == Medium.ANIME:
        return 'anime'
    elif medium == Medium.MANGA:
        return 'manga'
    else:
        raise ValueError('Only anime and managa are supported.')


def get_synonyms(entry: dict, site: Site):
    """
    Yield all synonyms from an entry.

    :param entry: the request data.

    :param site: The site for the entry.

    :return: A generator that yields all synonyms from a entry.
    """
    if not entry:
        return
    if site == Site.ANILIST:
        keys = ('title_english', 'title_romaji')
        lst_key = 'synonyms'
    elif site == Site.MAL:
        keys = ('title', 'english')
        lst_key = 'synonyms'
    elif site == Site.ANIDB:
        keys = ()
        lst_key = 'titles'
        pass
    else:
        return

    for key in keys:
        val = entry.get(key)
        if val:
            yield val.rstrip()

    lst = entry.get(lst_key, ())
    if lst and not isinstance(lst, str):
        for syn in lst:
            yield syn.strip()



from setuptools import setup, find_packages

with open('.version') as v:
    version = v.read()

with open('README.md') as r:
    readme = r.read()

with open('requirements.txt') as req:
    requirements = req.read().splitlines()

extras_require = {
    'postgres': ['asyncpg>=0.12.0']
}

package_data = {
    '': ['LICENSE', 'README.md', 'COPYING', 'requirements.txt', '.version']
}

setup(
    name='minoshiro',
    version=version,
    description=('An async Python3.6 library to search for anime, manga and'
                 'light novel using various web apis.'),
    long_description=readme,
    url='https://github.com/Mino-shiro/Minoshiro',
    author='MaT1g3R, dashwav',
    license='MIT',
    packages=find_packages(exclude=['tests', 'docs']),
    include_package_data=True,
    package_data=package_data,
    install_requires=requirements,
    extras_require=extras_require,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: AsyncIO',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English'
    ]
)



from enum import Enum

__all__ = ['Site', 'Medium']


class Site(Enum):
    MAL = 1
    ANILIST = 2
    ANIMEPLANET = 3
    ANIDB = 4
    KITSU = 5
    MANGAUPDATES = 6
    LNDB = 7
    NOVELUPDATES = 8
    VNDB = 9


class Medium(Enum):
    ANIME = 1
    MANGA = 2
    LN = 3
    VN = 4



from asyncio import get_event_loop
from json import dumps, loads
from pathlib import Path
from sqlite3 import connect
from time import time
from typing import Dict, Optional, Union

from minoshiro.enums import Medium, Site
from minoshiro.logger import get_default_logger
from minoshiro.upstream import get_all_synonyms
from .abc import DataController
from .constants import convert_medium, tables
from .sqlite_utils import make_tables


class SqliteController(DataController):
    """
    A SQLite3 data controller.
    """
    __slots__ = ('path', '_loop')

    def __init__(self, path: Union[str, Path], logger, loop=None):
        """
        Init method. Create the instance with the `get_instance` method to make
        sure you have all the tables needed.

        :param path:
            Path to the database,
            can be a string or a Pathlib Path object.

        :param logger: The logger object used for logging.

        :param loop:
            The asyncio event loop.
            If None is provided will use the default event loop.
        """
        self.path = str(path)
        self._loop = loop
        super().__init__(logger)

    @classmethod
    async def get_instance(cls, path: Union[str, Path], logger=None,
                           loop=None):
        """
        Get a new instance of `SqliteController`

        This method will create the appropriate tables needed.

       :param path:
            Path to the database,
            can be a string or a Pathlib Path object.

        :param logger: The logger object used for logging.

        :param loop:
            The asyncio event loop.
            If None is provided will use the default event loop.

        :return: A new instance of `SqliteController`
        """
        logger = logger or get_default_logger()
        logger.info('Creating tables...')
        await make_tables(path, loop or get_event_loop())
        logger.info('Tables created.')
        return cls(path, logger, loop)

    async def get_identifier(self, query: str,
                             medium: Medium) -> Optional[Dict[Site, str]]:
        """
        Get the identifier of a given search query.

        :param query: the search query.

        :param medium: the medium type.

        :return:
            A dict of all identifiers for this search query for all sites,
            None if nothing is found.
        """
        sql = """
        SELECT site, identifier FROM lookup
        WHERE LOWER(syname)=LOWER(?) AND medium=?
        """
        rows = await self.fetchall(sql, (query, medium.value))
        if not rows:
            return
        return {Site(site): id_ for site, id_ in rows if id_}

    async def set_identifier(self, name: str, medium: Medium,
                             site: Site, identifier: str):
        """
        Set the identifier for a given name.

        :param name: the name.

        :param medium: the medium type.

        :param site: the site.

        :param identifier: the identifier.
        """
        sql = 'REPLACE INTO lookup VALUES (?,?,?,?)'
        await self.execute(sql, (name, medium.value, site.value, identifier))

    async def get_mal_title(self, id_: str, medium: Medium) -> Optional[str]:
        """
        Get a MAL title by its id.

        :param id_: th MAL id.

        :param medium: the medium type.

        :return: The MAL title if it's found.
        """
        sql = 'SELECT title FROM mal WHERE id=? AND medium=?'
        row = await self.fetchone(sql, (id_, medium.value))
        if not row:
            return
        return row[0]

    async def set_mal_title(self, id_: str, medium: Medium, title: str):
        """
        Set the MAL title for a given id.

        :param id_: the MAL id.

        :param medium: The medium type.

        :param title: The MAL title for the given id.
        """
        sql = 'REPLACE INTO mal VALUES (?, ?, ?)'
        await self.execute(sql, (id_, medium.value, title))

    async def medium_data_by_id(self, id_: str, medium: Medium,
                                site: Site) -> Optional[dict]:
        """
        Get data by id.

        Note that if the data cache is more than 1 day old this will delete
        the row in the DB and return None.

        :param id_: the id.

        :param medium: the medium type.

        :param site: the site.

        :return: the data for that id if found.
        """
        sql = (f'SELECT dict, cachetime FROM {tables[medium]} '
               f'WHERE id=? AND site=?')
        row = await self.fetchone(sql, (id_, site.value))
        if not row:
            return
        data, cachetime = row
        now = int(time())
        if now - cachetime > 86400:
            await self.delete_medium_data(id_, medium, site)
            return
        return loads(data) if data else None

    async def set_medium_data(self, id_: str, medium: Medium,
                              site: Site, data: dict):
        """
        Set the data for a given id.

        :param id_: the id.

        :param medium: the medium type.

        :param site: the site.

        :param data: the data for the id.
        """
        sql = f'REPLACE INTO {tables[medium]} VALUES (?, ?, ?, ?)'
        await self.execute(sql, (id_, site.value, dumps(data), int(time())))

    async def delete_medium_data(self, id_: str, medium: Medium, site: Site):
        """
        Delete a row in medium data table.

        :param id_: the id.

        :param medium: the medium type.

        :param site: the site.
        """
        sql = f'DELETE FROM {tables[medium]} WHERE id=? AND site=?'
        try:
            await self.execute(sql, (id_, site.value))
        except Exception as e:
            self.logger.warning(str(e))

    async def pre_cache(self, session_manager):
        """
        Populate the lookup with synonyms.

        :param session_manager: The Aiohttp SessionManager.
        """
        rows = await get_all_synonyms(session_manager)
        with connect(self.path) as conn:
            for name, type_, db_links in rows:
                dict_ = loads(db_links)
                mal_name, mal_id = dict_.get('mal', ('', ''))
                anilist = dict_.get('ani')
                ap = dict_.get('ap')
                anidb = dict_.get('adb')
                medium = convert_medium[type_]
                if mal_name and mal_id:
                    _cache_mal(conn, str(mal_id), medium, str(mal_name))
                _precache(conn, name, medium, Site.MAL, mal_id)
                _precache(conn, name, medium, Site.ANILIST, anilist)
                _precache(conn, name, medium, Site.ANIMEPLANET, ap)
                _precache(conn, name, medium, Site.ANIDB, anidb)
            conn.commit()

    @property
    def loop(self):
        """
        :return: `self._loop` or a default event loop.
        """
        return self._loop or get_event_loop()

    def __execute(self, sql: str, params=None):
        """
        Execute and commit and SQL query.

        :param sql: the SQL query.

        :param params: the SQL parameters.
        """
        with connect(self.path) as conn:
            conn.execute(sql, params)
            conn.commit()

    async def execute(self, sql: str, params=None):
        """
        Run `self.__execute` using an asyncio event loop.

        :param sql: the SQL query.

        :param params: the SQL parameters.
        """
        await self.loop.run_in_executor(
            None, self.__execute, sql, params
        )

    def __fetch(self, all_: bool, sql: str, params=None):
        """
        Fetch results from a SQL query.

        :param all_: True to fetch all rows, False to fetch one row.

        :param sql: the SQL query.

        :param params: the SQL parameters.

        :return: The results fetched from the SQL query.
        """
        with connect(self.path) as conn:
            cur = conn.execute(sql, params)
            if all_:
                return cur.fetchall()
            else:
                return cur.fetchone()

    async def fetchall(self, sql: str, params=None):
        """
        Run `self.__fetch` using an asyncio event loop.

        :param sql: the SQL query.

        :param params: the SQL parameters.

        :return: All rows fetched from the SQL query.
        """
        return await self.loop.run_in_executor(
            None, self.__fetch, True, sql, params
        )

    async def fetchone(self, sql: str, params=None):
        """
        Run `self.__fetch` using an asyncio event loop.

        :param sql: the SQL query.

        :param params: the SQL parameters.

        :return: One row fetched from the SQL query.
        """
        return await self.loop.run_in_executor(
            None, self.__fetch, False, sql, params
        )


def _precache(conn, name, medium, site, id_):
    """
    Cache id.
    """
    if name and (id_ or isinstance(id_, int)):
        sql = 'REPLACE INTO lookup VALUES (?,?,?,?)'
        conn.execute(
            sql, (str(name), medium.value, site.value, str(id_))
        )


def _cache_mal(conn, id_: str, medium: Medium, title: str):
    """
    Cache mal title.
    """
    sql = 'REPLACE INTO mal VALUES (?, ?, ?)'
    conn.execute(sql, (id_, medium.value, title))



from .abc import DataController
from .postgres_controller import PostgresController
from .sqlite_controller import SqliteController

__all__ = ['PostgresController', 'DataController', 'SqliteController']
